﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using xserver;

namespace Geometry
{
    /**
     * Implementation of the Bowyer-Watson algorithm (http://en.wikipedia.org/wiki/Bowyer%E2%80%93Watson_algorithm)
     * Calculates the delaunay-triangulation of a set of locations
    **/
    class BowyerWatsonAlgorithm
    {
        private Triangulation delaunayTriangulation { get; set; }

        public BowyerWatsonAlgorithm()
        {
            delaunayTriangulation = new Triangulation();
        }
        
        public Triangulation calculateDelaunay(List<Location> locations)
        {
            // Fill list with points from the input locations
            List<Point> points = new List<Point>();
            for (int id = 0; id < locations.Count; id++)
            {
                points.Add(new Point(locations[id].coordinate.point, Convert.ToInt32(locations[id].id)));
            }

            /**
             * Calculate a triangle, which encloses all input locations:
             * This is done by constructing a triangle around the bounding box of all points
            **/
            double minY = Double.MaxValue, maxY = Double.MinValue, minX = Double.MaxValue, maxX = Double.MinValue;
            foreach (Point p in points)
            {
                if (p.pt.x < minX)
                    minX = p.pt.x;
                if (p.pt.x > maxX)
                    maxX = p.pt.x;
                if (p.pt.y < minY)
                    minY = p.pt.y;
                if (p.pt.y > maxY)
                    maxY = p.pt.y;
            }

            double dy = maxY - minY;
            double dx = maxX - minX;

            Point A = new Point(new PlainPoint { x = minX - dy, y = minY }, -1);
            Point B = new Point(new PlainPoint { x = maxX + dy, y = minY }, -2);
            Point C = new Point(new PlainPoint { x = minX + 0.5 * dx, y = maxY + 0.5 * dx }, -3);

            // Create triangle around bounding box and enlarge it with a scale factor, sothat no points are on the edges
            PlainPoint center = new PlainPoint { x = minX + 0.5 * dx, y = minY + 0.5 * dy };
            Triangle startingTriangle = new Triangle(A, B, C, 0);
            startingTriangle.scale(10, center);

            // Add initial triangle to the delaunay triangulation
            delaunayTriangulation.addTraingle(startingTriangle);

            // Start calculation of the delaunay triangulation
            foreach (Point p in points)
            {
                // Get all triangles, which are no longer valid, due to insertion of p
                List<Triangle> badTriangles = delaunayTriangulation.getBadTriangles(p);

                // Get polygonal hole as a list of edges
                List<Edge> polygon = new List<Edge>();
                foreach (Triangle tri in badTriangles)
                {
                    foreach (Edge e in tri.edges)
                    {
                        bool shared = false;
                        foreach (Triangle otherTri in badTriangles)
                        {
                            if (tri.ID != otherTri.ID)
                            {
                                if (otherTri.containsEdge(e))
                                {
                                    shared = true;
                                }
                            }
                        }
                        if (!shared)
                        {
                            polygon.Add(e);
                        }
                    }
                }
                // Remove all bad triangles from the triangulation
                delaunayTriangulation.removeBadTriangles();
                // Retriangulate the polygonal hole
                delaunayTriangulation.reTriangulate(polygon, p);
            }
            // Remove the initial triangle and all triangles connected to it
            delaunayTriangulation.removeStartingTriangle(startingTriangle);
            return delaunayTriangulation;
        }

        private bool checkDelaunay(List<Point> points)
        {
            PointTest pt = new PointTest();
            foreach (Point p in points)
            {
                foreach (Triangle t in delaunayTriangulation.triangles)
                {
                    if (t.isBad(p))
                    {
                        if (!(pt.areEqual(p, t.A) || pt.areEqual(p, t.B) || pt.areEqual(p, t.C)))
                        {
                            return false;
                        }
                    }
                }
            }
            return true;
        }
    }
}
