﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using xserver;

namespace Geometry
{
    /**
     * Represents a triangulation, given as a set of triangles
    **/
    class Triangulation
    {
        public List<Triangle> triangles { get; private set; }

        public Triangulation()
        {
            triangles = new List<Triangle>();
        }

        // Adds a triangle to the triangulation
        public void addTraingle(Triangle tri)
        {
            triangles.Add(tri);
        }

        /**
         * Fetches a list of all Triangles, which are no longer valid (with respect to delaunay properties), when p is inserted 
         * I.e. all triangles, for which p is inside the circumcircle are returned
         * Also marks the triangles as bad, for future removal
        **/
        public List<Triangle> getBadTriangles(Point p)
        {
            List<Triangle> badTriangles = new List<Triangle>();
            foreach (Triangle tri in triangles)
            {
                if (tri.isBad(p))
                {
                    badTriangles.Add(tri);
                    tri.bad = true;
                }
            }
            return badTriangles;
        }

        // Removes all bad triangles
        public void removeBadTriangles()
        {
            int iterator = 0;
            while (triangles.Count != 0 && iterator < triangles.Count)
            {
                if (triangles[iterator].bad)
                    triangles.RemoveAt(iterator);
                else
                    iterator++;
            }
        }

        /**
         * Retriangulates the polygonal hole, given as a list of edges
         * All edges are connected with the point p, to form a triangle
        **/
        public void reTriangulate(List<Edge> edges, Point p)
        {
            foreach (Edge e in edges)
            {
                int id = 0;
                if (triangles.Count != 0)
                    id = triangles[triangles.Count - 1].ID + 1;
                triangles.Add(new Triangle(e.P, e.Q, p, id));
            }
        }

        // Removes all triangles, which share a node with the initial triangle
        public void removeStartingTriangle(Triangle startingTriangle)
        {
            int iterator = 0;
            while (triangles.Count != 0 && iterator < triangles.Count)
            {
                if (triangles[iterator].containsNodeFrom(startingTriangle))
                {
                    triangles.RemoveAt(iterator);
                }
                else
                {
                    iterator++;
                }
            }
        }
    }
}
