﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using xserver;

namespace VisualizingTerritoriesAsAreas
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        public MainWindow()
        {
            InitializeComponent();
            runSample();
        }

        /**
         * This sample will use PTV xTerritory, to calculate 3 territories for a set of locations with one activity measure.
         * The territories will subsequently be visualized, using PTV xMap. 
         * In this sample the convex hulls of the territories will be used for visualization
        **/
        private void runSample()
        {
            Territories territories = calculateTerritories();
            if (territories != null)
                mapTerritories(ref territories);
        }

        private Territories calculateTerritories()
        {
            //Step 1: Initiate xTerritory client
            XTerritoryWSService xTerritoryClient = new XTerritoryWSService();

            //Step 2: Create CallerContext with appropriate profile and coordinate format
            CallerContext context = new CallerContext();
            CallerContextProperty props1 = new CallerContextProperty
            {
                key = "Profile",
                value = "default"
            };
            CallerContextProperty props2 = new CallerContextProperty
            {
                key = "CoordFormat",
                value = "PTV_MERCATOR"
            };
            CallerContextProperty[] properties = new CallerContextProperty[] { props1, props2 };
            context.wrappedProperties = properties;

            //Step 3: Load the locations for territory planning
            List<String> locationStrings = getLocations();
            Location[] locations = new Location[locationStrings.Count];
            for (int i = 0; i < locationStrings.Count; i++)
            {
                String[] splitString = locationStrings[i].Split(';');
                String locID = splitString[0];

                //Step 3.1: Load x and y coordinates as well as activity measure
                double locX = Convert.ToDouble(splitString[1]);
                double locY = Convert.ToDouble(splitString[2]);
                double locActivity = Convert.ToDouble(splitString[3]);

                xserver.Point point = new xserver.Point
                {
                    point = new PlainPoint { x = locX, y = locY }
                };

                //Step 3.2: Create a Location with the data above
                locations[i] = new Location
                {
                    id = locID,
                    coordinate = point,
                    activity = new AbsoluteActivity { value = locActivity }
                };
            }

            //Step 4: Create DistanceCalculationOptions, here a new distance matrix will be used.
            ApproximateByDirectDistance distOptions = new ApproximateByDirectDistance();

            //Step 5: Create PlanningOptions and specify the number of territories
            PlanningOptions options = new PlanningOptions
            {
                exactNumberOfTerritories = 3                //the visualization part of this sample supports a maximum of 3 territories
            };

            //Step 6: Create PlanningProfile
            PlanningProfile profile = new PlanningProfile
            {
                solutionQuality = SolutionQuality.BASIC,    //Solution quality, with larger problems this impacts the running time severely
                avoidOverlappingTerritories = true,         //If true, the algorithm tries to avoid visual overlaps between territories,
                                                            //if a solution without overlaps is possible
                createBalancedTerritories = true,           //If true, the imbalanceTolerance parameter will be considered
                imbalanceTolerance = 10.0,                  //Maximum deviation from average activity per territory
            };

            //Step 7: Create ResponseContents, specify which information the client returns
            ResponseContents contents = new ResponseContents
            {
                locations = true,
                territories = true,
            };

            try
            {
                //Step 8: Calculate territories
                Territories territories = xTerritoryClient.planTerritories(locations, distOptions, options, profile, contents, context);
                return territories;
            }
            catch (Exception e)
            {
                return null;
            }
        }

        private void mapTerritories(ref Territories territories)
        {
            //Step 1: Create xMap client
            XMapWSService client = new XMapWSService();

            //Step 2: Create ImageInfo
            ImageInfo imageInfo = new ImageInfo
            {
                format = ImageFileFormat.PNG,
                height = 800,
                width = 1280,
                imageParameter = ""
            };

            //Step 3: Create MapParams
            MapParams mapParams = new MapParams
            {
                showScale = true,
                useMiles = false
            };

            //Step 4.1: Create Bitmap array for locations, each location will be visualized with a bitmap marker
            Bitmap[] bitmapArrayLocations = new Bitmap[territories.wrappedLocations.Length];
            
            //Step 4.2: Create Bitmap array for territory centers, each center will be visualized with a bitmap marker
            Bitmap[] bitmapArrayTerritories = new Bitmap[territories.wrappedTerritories.Length];
            
            int locationCounter = 0;
            int territoryCounter = 0;

            //Step 5: Create list of territoryIDs, it will be used to match with the territoryIDs of the locations
            List<String> territoryIDs = new List<String>();

            //Step 6: Create Dictionary with territoryID as ID and lists of locations as entries
            //        For each territory all locations will be saved, which it contains
            Dictionary<String, List<Location>> locationLists = new Dictionary<String, List<Location>>();
            foreach (Territory ter in territories.wrappedTerritories)
            {
                locationLists.Add(ter.id, new List<Location>());
            }
            foreach (Location loc in territories.wrappedLocations)
            {
                locationLists[loc.territoryId].Add(loc);
            }

            Geometry.GrahamScan gs = new Geometry.GrahamScan();
            List<GeometryLayer> geometryLayers = new List<GeometryLayer>();
            foreach (Territory ter in territories.wrappedTerritories)
            {
                territoryIDs.Add(ter.id);

                //Step 8: In this step the convex hull of the district will be calculated and added to a geometry layer
                //Step 8.1: Calculate convex hull
                List<Location> hullLocations = gs.calculateConvexHull(locationLists[ter.id]);

                //Step 8.2: Add locations on the hull into a PlainPoint array
                PlainPoint[] plainPoints = new PlainPoint[hullLocations.Count + 1];
                int index = 0;
                foreach (Location loc in hullLocations)
                {
                    plainPoints[index] = loc.coordinate.point;
                    index++;
                }
                
                //Step 8.3: Add the first point at the end again
                plainPoints[plainPoints.Length - 1] = hullLocations[0].coordinate.point;

                //Step 8.4: Create a polygon from the plainpoints
                xserver.Polygon poly = new xserver.Polygon { 
                    polygon = new PlainPolygon { 
                        wrappedLinearRings = new PlainLinearRing[] { 
                            new PlainLinearRing { wrappedPoints = plainPoints } 
                        } 
                    } 
                };

                //Step 8.5: Add the polygon to a Geometry object
                xserver.Geometry hullPolygon = new xserver.Geometry
                {
                    description = "hullPolygon",
                    id = territoryCounter,
                    geometry = poly,
                    referencePoint = ter.referencePoint
                };

                //Step 8.5: Determine the appropriate color for this territory
                String color = rgbToColorString(0, 255, 0);
                if (territoryCounter == 1)
                    color = rgbToColorString(255, 0, 0);
                if (territoryCounter == 2)
                    color = rgbToColorString(0, 0, 255);
                
                //Step 8.6: Add geometry options
                GeometryOption[] geometryOptions = new GeometryOption[]{  
                    new GeometryOption { option = GeometryOptions.BORDERCOLOR,   
                                                value = rgbToColorString(255,255,255)},  
                    new GeometryOption { option = GeometryOptions.FILLCOLOR,   
                                                value = color},  
                    new GeometryOption { option = GeometryOptions.FILLALPHA,   
                                                value = "215"}, // 0-255  
                    new GeometryOption { option = GeometryOptions.AUTOCENTEROBJECTS,   
                                                value = "1"} // 0-255  
                };

                //Step 8.7: Add the polygon and the options into a geometries array
                xserver.Geometries[] geometries = new xserver.Geometries[]   
                {  
                    new xserver.Geometries {  
                        wrappedGeometries = new xserver.Geometry[] { hullPolygon },  
                        wrappedOptions = geometryOptions,  
                    }  
                };

                //Step 8.8: Create a geometry layer and add it to the list
                GeometryLayer geomLayer = new GeometryLayer
                {
                    visible = true,
                    wrappedGeometries = geometries,
                    drawPriority = 25000,
                    objectInfos = ObjectInfoType.GEOMETRY,
                    name = ter.id,
                };
                geometryLayers.Add(geomLayer);

                territoryCounter++;
            }

            //Step 9: Add locations to bitmap array
            foreach (Location location in territories.wrappedLocations)
            {
                String bmp = "location.bmp";
                if (location.territoryId == territoryIDs[0])
                {
                    bmp = "flaggreen.bmp";
                }
                else if (location.territoryId == territoryIDs[1])
                {
                    bmp = "flagred.bmp";
                }
                else if (location.territoryId == territoryIDs[2])
                {
                    bmp = "flagblue.bmp";
                }
                xserver.Point point = location.coordinate;
                bitmapArrayLocations[locationCounter] = new Bitmap
                {
                    name = bmp,
                    position = point
                };

                locationCounter++;
            }

            //Step 10: Create BitmapOptions and Bitmaps arrays
            BitmapOptions bitmapOptionsLocations = new BitmapOptions
            {
                referencePoint = new PlainPoint { x = 25, y = 32 },
                transparencyColor = new xserver.Color { red = 255, green = 255, blue = 255 }
            };
            Bitmaps[] bitmaps = new Bitmaps[] 
            { 
                new Bitmaps { wrappedBitmaps = bitmapArrayLocations, options = bitmapOptionsLocations },
            };

            //Step 11: Create CustomLayer for bitmaps
            CustomLayer customBMPLayer = new CustomLayer
            {
                centerObjects = true,
                drawPriority = 25002,
                name = "ObjectLayer",
                objectInfos = ObjectInfoType.REFERENCEPOINT,
                visible = true,
                wrappedBitmaps = bitmaps
            };

            //Step 12: Add all CustomLayers (bitmaps and geometry) to a Layer array
            Layer[] layers = new Layer[geometryLayers.Count + 1];
            layers[0] = customBMPLayer;
            for (int i = 1;  i < layers.Length; i++)
            {
                layers[i] = geometryLayers[i - 1];
            }

            bool includeImageInResponse = false;

            try
            {
                // Step 13: Return map object with its characteristics
                Map map = client.renderMap(null, mapParams, imageInfo, layers, includeImageInResponse, null);

                // Step 15: Retrieve image
                String imageUrl = "http://" + map.image.url;
                // Create BitmapImage
                BitmapImage bitmapImage = new BitmapImage();

                // Display BitmapImage
                bitmapImage.BeginInit();
                bitmapImage.UriSource = new Uri(imageUrl);
                bitmapImage.EndInit();

                img1.Source = bitmapImage;

            }
            catch (Exception e)
            {
                // proper exception handling: see API for PTV's exception handling!
            }
        }

        private String rgbToColorString(int red, int green, int blue)
        {
            int color = (blue << 16) + (green << 8) + (red);
            return "" + color;
        }  
        
        private List<String> getLocations()
        {
            List<String> list = new List<String>();

            // orderId, X, Y, activity
            list.Add("101;671006;6382294;90");
            list.Add("102;681812;6375071;120");
            list.Add("103;682676;6388154;90");
            list.Add("104;675000;6380000;60");
            list.Add("105;679143;6374455;90");
            list.Add("106;684621;6374703;180");
            list.Add("107;684659;6372947;90");
            list.Add("108;683151;6371248;30");
            list.Add("109;685669;6371954;45");
            list.Add("110;683974;6369193;60");
            list.Add("111;683000;6366903;90");
            list.Add("112;678763;6367800;120");
            list.Add("113;681397;6369670;105");
            list.Add("114;678400;6370854;90");
            list.Add("115;674779;6373188;60");
            list.Add("116;678149;6380180;45");
            list.Add("117;680440;6381326;30");
            list.Add("118;665627;6378195;90");
            list.Add("119;687472;6386122;180");
            list.Add("120;666773;6381020;120");
            list.Add("121;687136;6382563;60");
            list.Add("122;665642;6382471;30");
            list.Add("123;665627;6386365;45");
            list.Add("124;672957;6388121;60");
            list.Add("125;677004;6388121;75");
            list.Add("126;682578;6384074;90");
            list.Add("127;687540;6379111;105");
            list.Add("128;681661;6384838;120");
            list.Add("129;687388;6385754;135");
            list.Add("130;685937;6388121;150");

            return list;
        }
    }
}
